// *********************************************************************
//			NList Class
// Here we define the functions for the class NList. All you need to
// manipulate a List of Points. This version uses two pointers, one to
// the previous node and the other to the next node (double linked list).
// A particular node named "sentinel" indicates the end of list.
// Initial version:	09/12/97
// *********************************************************************
//   $Id: NList.H,v 1.3 1999/09/09 14:49:16 avillani Exp $



#ifndef NLIST_HDR
#define NLIST_HDR

#include "GenException.H"


// Base Classes

class ListBaseNode {
  friend class BaseNList;   // NList can access the private memebers of the
                            // ListBaseNode Class
  ListBaseNode* prev;	    // Pointer to the previous node
  ListBaseNode* next;	    // Pointer to the next node
};


typedef ListBaseNode* NListPos;	// A position in the List is a Node pointer

class BaseNList {
  ListBaseNode* sentinel;	// Pointer to the sentinel
  ListBaseNode* current;	// Pointer to the current node
  int len;			// Number of items in the list

public:
  BaseNList();			// Empty constructor
  ~BaseNList();			// Destructor

  inline NListPos Next();		// Move current to the next node
  inline NListPos Previous();		// Move current to the previous node
  NListPos GoTo(NListPos p)	// Move current to the node pointed to by p
    { return(current = p); }
  NListPos GoToFirst()		// Move current to the first node
    { return(current = sentinel->next); }
  NListPos GoToLast()		// Move current to the last node
    { return(current = sentinel->prev); }
  NListPos GoToEndOfList()	// Move current to the sentinel (end of list)
    { return(current = sentinel); }

  int IsEmpty()	const			  // The List is empty?
    { return(sentinel->next == sentinel); }
  int IsEndOfList() const		  // Current is on the sentinel?
    { return(current == sentinel); }
  int IsFirst() const			  // Current is the first item?
    { return(current == sentinel->next); }

  NListPos Insert(ListBaseNode* d);	  // Insert a "link"  before current
  NListPos InsertFirst(ListBaseNode* d);  // Insert a "link" before the first
                                          // node
  NListPos InsertLast(ListBaseNode* d);	  // Insert a "link" beyond the last 
                                          // node
  ListBaseNode* Delete();		  // Remove the node pointed to by
                                          // current and return it 

  NListPos Clear();	                  // Empty list but not destroy it
  NListPos Current() const		  // Return current
    { return(current); }

  int Size() const		          // Return number of items in the list
    { return(len); }
};


// **************************************************************************
// Derived Classes
  
template<class MyType> class NList;    // Prototype for "forward" definition

// The Node class expands the ListBaseNode class, providing a data field of
// type MyNType
template<class MyNType>class Node: private ListBaseNode {
  friend class NList<MyNType>;	    // NList<MyNType> can access the private 
                                    // member data of Node class

  MyNType data;

public:
  Node(const MyNType& d):           // Copy constructor
    data(d)
    {}
};


// The NList class expands the BaseNList class, giving the right implementation
// of the NList class functions.
template<class MyType>class NList: public BaseNList {
public:
  ~NList();

  NListPos Insert(const MyType &d);
  NListPos InsertFirst(const MyType &d);	// Add at the head of the List
  NListPos InsertLast(const MyType &d);		// Add at the tail of the List

  MyType Delete();				// Remove the current item from
						// the list (and return it)

  NListPos Write(const MyType &d);	// Write a new data in the current item
  MyType Read() const;			// Return the data of the current item

  NListPos Clear();
};


// The CrList (Circular List) class expands the NList class, redefining the 
// Next() and Previous() functions
template<class MyType>class CrList: public NList<MyType> {
public:
  ~CrList();
  inline NListPos Next();
  inline NListPos Previous();
};


// **************************************************************************
// Template class NList functions

// Destructor
template<class MyType> NList<MyType>::~NList()
{
  Clear();
}


// Insert a node containing the d data before current
template<class MyType> NListPos NList<MyType>::Insert(const MyType& d)
{
  // Build an object of type Node<MyType>
  Node<MyType> *nptr = new Node<MyType>(d);

  return(BaseNList::Insert(nptr));
}


// Insert a node containing the d data before the first node
template<class MyType> NListPos NList<MyType>::InsertFirst(const MyType& d)
{
  GoToFirst();

  return(Insert(d));
}


// Insert a node containing the d data beyond the last node
template<class MyType> NListPos NList<MyType>::InsertLast(const MyType& d)
{
  GoToEndOfList();

  return(Insert(d));
}


 // Return data of current
template<class MyType> MyType NList<MyType>::Read() const
{
  if (IsEndOfList()) {
    GenericException theexcept("NList<MyType>::Read",
			       "You can't read the sentinel value.");

    // Throw the exception
    throw (theexcept);
  }

  Node<MyType>* npntr = (Node<MyType>*) BaseNList::Current();

  return(npntr->data);
}


// Write a new d data into current and return current
template<class MyType> NListPos NList<MyType>::Write(const MyType& d)
{
  if (IsEndOfList()) {
    GenericException theexcept("NList<MyType>::Write",
			       "You can't write on the sentinel.");

    // Throw the exception
    throw (theexcept);
  }

  ListBaseNode* currnt = BaseNList::Current();
  ((Node<MyType>*) currnt)->data = d;

  return(currnt);
}


// Remove current and return its data field.
template<class MyType> MyType NList<MyType>::Delete()
{
  Node<MyType> *npntr = (Node<MyType>*) BaseNList::Delete();

  MyType tmp(npntr->data);
  delete npntr;

  return(tmp);
}


 // Empty list but not destroy it.
template<class MyType> NListPos NList<MyType>::Clear()
{
  GoToFirst();			

  while (!IsEndOfList())	
    Delete();			

  return(Current());
}


// Template class CrList functions

// Destructor
template<class MyType> CrList<MyType>::~CrList()
{
  Clear();
}


// Move current to the next node (only if the list is not empty). If current 
// is at end of list or on the last node, current is moved to the first node;
// otherwise BaseNList::Next() is used. Finally the value of current is 
// returned.  
template<class MyType> NListPos CrList<MyType>::Next()
{
  if (!IsEmpty()) {
    if (IsEndOfList())
      GoToFirst();
    else {
      BaseNList::Next();
      if (IsEndOfList())
        GoToFirst();
    }
  }
  return (Current());
}


// Move current to the previous node (only if the list is not empty). If 
// current is on the first node, current is moved on the last node. Otherwise
// BaseNList::Previous() is used. Finally the value of current is returned.
template<class MyType> NListPos CrList<MyType>::Previous()
{
  if (!IsEmpty()) {
    if (IsFirst())
      GoToLast();
    else
      BaseNList::Previous();
  }
  return (Current());
}   

// Move current to the next node
NListPos BaseNList::Next()
{
  if (IsEndOfList()) {
    GenericException theexcept("BaseNList::Next",
			       "You can't go beyond the sentinel.");
    // Throw the exception
    throw (theexcept);
  }

  return(current = current->next);
}


// Move current to the previous node
NListPos BaseNList::Previous()
{
  if (IsFirst()) {
    GenericException theexcept("BaseNList::Previous",
			       "You can't jump on the sentinel.");

    // Throw the exception
    throw (theexcept);
  }

  return(current = current->prev);
}

#endif
